# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division, print_function)

from .numbers import number_to_scientific_html, number_to_scientific_latex, number_to_scientific_unicode
from .table import as_per_substance_html_table
from .js import javascript
from .string import str_
from .pretty import unicode_
from .web import html, css
from .tex import latex
